@@ -20,6 +20,7 @@ import (
 	"path/filepath"
 	"runtime"
 	"strings"
+	"sync"
 	"time"
 
 	"github.com/cloudflare/cfrpki/sync/lib"
@@ -32,14 +33,13 @@ import (
 	"github.com/prometheus/client_golang/prometheus/promhttp"
 
 	"github.com/cloudflare/gortr/prefixfile"
-	"github.com/gorilla/mux"
 	log "github.com/sirupsen/logrus"
-)
 
-const (
-	RRDP_NO_MATCH = iota
-	RRDP_MATCH_RSYNC
-	RRDP_MATCH_STRICT
+	// Debugging
+	"github.com/getsentry/sentry-go"
+	"github.com/opentracing/opentracing-go"
+	jcfg "github.com/uber/jaeger-client-go/config"
+	"net/http/pprof"
 )
 
 var (
@@ -55,24 +55,29 @@ var (
 	LogLevel    = flag.String("loglevel", "info", "Log level")
 	Refresh     = flag.String("refresh", "20m", "Revalidation interval")
 
+	StrictManifests = flag.Bool("strict.manifests", true, "Manifests must be complete or invalidate CA")
+	StrictHash      = flag.Bool("strict.hash", true, "Check the hash of files")
+	StrictCms       = flag.Bool("strict.cms", false, "Decode CMS with strict settings")
+
 	// Rsync Options
 	RsyncTimeout = flag.String("rsync.timeout", "20m", "Rsync command timeout")
 	RsyncBin     = flag.String("rsync.bin", DefaultBin(), "The rsync binary to use")
 
 	// RRDP Options
-	RRDP     = flag.Bool("rrdp", true, "Enable RRDP fetching")
-	RRDPFile = flag.String("rrdp.file", "rrdp.json", "Save RRDP state")
-	RRDPMode = flag.Int("rrdp.mode", RRDP_MATCH_RSYNC, fmt.Sprintf("RRDP security mode (%v = no check, %v = match rsync domain, %v = match path)",
-		RRDP_NO_MATCH, RRDP_MATCH_RSYNC, RRDP_MATCH_STRICT))
+	RRDP         = flag.Bool("rrdp", true, "Enable RRDP fetching")
+	RRDPFile     = flag.String("rrdp.file", "cache/rrdp.json", "Save RRDP state")
+	RRDPFailover = flag.Bool("rrdp.failover", true, "Failover to rsync when RRDP fails")
+	UserAgent    = flag.String("useragent", fmt.Sprintf("Cloudflare-RRDP-%v (+https://github.com/cloudflare/cfrpki)", AppVersion), "User-Agent header")
 
 	Mode       = flag.String("mode", "server", "Select output mode (server/oneoff)")
 	WaitStable = flag.Bool("output.wait", true, "Wait until stable state to create the file (returns 503 when unstable on HTTP)")
 
 	// Serving Options
-	Addr        = flag.String("http.addr", ":8080", "Listening address")
+	Addr        = flag.String("http.addr", ":8081", "Listening address")
 	CacheHeader = flag.Bool("http.cache", true, "Enable cache header")
 	MetricsPath = flag.String("http.metrics", "/metrics", "Prometheus metrics endpoint")
 	InfoPath    = flag.String("http.info", "/infos", "Information URL")
+	HealthPath  = flag.String("http.health", "/health", "Health URL")
 
 	CorsOrigins = flag.String("cors.origins", "*", "Cors origins separated by comma")
 	CorsCreds   = flag.Bool("cors.creds", false, "Cors enable credentials")
@@ -83,6 +88,11 @@ var (
 	SignKey  = flag.String("output.sign.key", "private.pem", "ECDSA signing key")
 	Validity = flag.String("output.sign.validity", "1h", "Validity")
 
+	// Debugging options
+	Pprof     = flag.Bool("pprof", false, "Enable pprof endpoint")
+	Tracer    = flag.Bool("tracer", false, "Enable tracer")
+	SentryDSN = flag.String("sentry.dsn", "", "Send errors to Sentry")
+
 	Version = flag.Bool("version", false, "Print version")
 
 	CertRepository = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 48, 5}
@@ -170,9 +180,14 @@ type RRDPInfo struct {
 	Serial    int64  `json:"serial"`
 }
 
+var errKeyNotParsed = fmt.Errorf("Failed to PEM decode key")
+
 func ReadKey(key []byte, isPem bool) (*ecdsa.PrivateKey, error) {
 	if isPem {
 		block, _ := pem.Decode(key)
+		if block == nil {
+			return nil, errKeyNotParsed
+		}
 		key = block.Bytes
 	}
 
@@ -203,13 +218,13 @@ type Stats struct {
 type state struct {
 	Basepath     string
 	Tals         []*pki.PKIFile
+	TalsFetch    map[string]*librpki.RPKITAL
 	TalNames     []string
 	UseManifest  bool
 	RsyncBin     string
 	RsyncTimeout time.Duration
 
-	Mode     string
-	RRDPMode int
+	Mode string
 
 	Validity     time.Duration
 	LastComputed time.Time
@@ -218,20 +233,23 @@ type state struct {
 	Key          *ecdsa.PrivateKey
 	EnableCache  bool
 
-	Stable      bool // Indicates something has been added to the fetch list (rsync of rrdp)
-	Fetcher     *syncpki.LocalFetch
-	HTTPFetcher *syncpki.HTTPFetcher
+	Stable            bool // Indicates something has been added to the fetch list (rsync of rrdp)
+	HasPreviousStable bool
+	Fetcher           *syncpki.LocalFetch
+	HTTPFetcher       *syncpki.HTTPFetcher
 
-	RsyncFetch    map[string]time.Time
-	RRDPFetch     map[string][]string
-	FailoverRsync []string
+	PrevRepos    map[string]time.Time
+	CurrentRepos map[string]time.Time
 
-	FinalRsyncFetch map[string]bool
-	FinalRRDPFetch  map[string][]string
+	RsyncFetch      map[string]string
+	RRDPFetch       map[string]string
+	RRDPFetchDomain map[string]string
 
-	RRDPInfo map[string]RRDPInfo
+	RRDPInfo     map[string]RRDPInfo
+	RRDPFailover bool
 
-	ROAList *prefixfile.ROAList
+	ROAList     *prefixfile.ROAList
+	ROAListLock *sync.RWMutex
 
 	// Various counters and statistics
 	RRDPStats          map[string]Stats
@@ -241,61 +259,36 @@ type state struct {
 	Iteration          int
 	ValidationMessages []string
 	ROAsTALsCount      []ROAsTAL
+
+	InfoAuthorities     [][]SIA
+	InfoAuthoritiesLock *sync.RWMutex
+
+	Pprof bool
+
+	StrictHash      bool
+	StrictManifests bool
+	StrictCms       bool
 }
 
 func (s *state) MainReduce() bool {
-	previousRsyncFetch := s.FinalRsyncFetch
-	previousRRDPFetch := s.FinalRRDPFetch
-
-	s.FinalRsyncFetch = make(map[string]bool)
-	s.FinalRRDPFetch = make(map[string][]string)
-
-	rsyncMap := make(map[string]syncpki.SubMap)
-	for k, _ := range s.RsyncFetch {
-		syncpki.AddInMap(k, rsyncMap)
-	}
-	rsyncRedMap := syncpki.ReduceMap(rsyncMap)
-
-	for _, v := range rsyncRedMap {
-		s.FinalRsyncFetch[v] = true
-	}
-
-	for k, v := range s.RRDPFetch {
-		rsyncMap = make(map[string]syncpki.SubMap)
-		for _, vv := range v {
-			syncpki.AddInMap(vv, rsyncMap)
-		}
-		rsyncRedMap = syncpki.ReduceMap(rsyncMap)
-		for _, vv := range rsyncRedMap {
-			if _, ok := s.FinalRRDPFetch[k]; !ok {
-				s.FinalRRDPFetch[vv] = make([]string, 0)
-			}
-
-			s.FinalRRDPFetch[vv] = append(s.FinalRRDPFetch[vv], k)
-
-			if _, ok := s.FinalRsyncFetch[vv]; ok {
-				log.Debugf("Deleting %v from rsync because there is an rrdp\n", vv)
-				delete(s.FinalRsyncFetch, vv)
-			}
+	var hasChanged bool
+	for rsync, ts := range s.CurrentRepos {
+		if _, ok := s.PrevRepos[rsync]; !ok {
+			s.PrevRepos[rsync] = ts
+			hasChanged = true
+			log.Debugf("Repository %s has appeared at %v", rsync, ts)
 		}
 	}
 
-	if len(s.FinalRRDPFetch) != len(previousRRDPFetch) ||
-		len(s.FinalRsyncFetch) != len(previousRsyncFetch) {
-		return true
-	}
-	for v, _ := range s.FinalRsyncFetch {
-		if _, ok := previousRsyncFetch[v]; !ok {
-			return true
-		}
-	}
-	for v, _ := range s.FinalRRDPFetch {
-		if _, ok := previousRRDPFetch[v]; !ok {
-			return true
-		}
+	// Init deletion of folder if missing from current
+
+	s.Fetcher.SetRepositories(s.CurrentRepos)
+
+	if !hasChanged && len(s.PrevRepos) != len(s.CurrentRepos) {
+		hasChanged = true
 	}
 
-	return false
+	return hasChanged
 }
 
 func ExtractRsyncDomain(rsync string) (string, error) {
@@ -307,20 +300,7 @@ func ExtractRsyncDomain(rsync string) (string, error) {
 	}
 }
 
-func (s *state) ReceiveRRDPFileCallback(main string, url string, path string, data []byte, withdraw bool, snapshot bool, serial int64, args ...interface{}) error {
-	rsync, _ := args[0].(string)
-	if s.RRDPMode == RRDP_MATCH_STRICT && !strings.Contains(path, rsync) {
-		log.Errorf("%v is outside directory %v", path, rsync)
-		return nil
-	}
-	if s.RRDPMode == RRDP_MATCH_RSYNC {
-		newDom, err := ExtractRsyncDomain(rsync)
-		if err == nil && !strings.Contains(path, newDom) {
-			log.Errorf("%v is outside directory %v", path, newDom)
-			return nil
-		}
-	}
-
+func (s *state) WriteRsyncFileOnDisk(path string, data []byte, withdraw bool) error {
 	fPath, err := syncpki.GetDownloadPath(path, true)
 	if err != nil {
 		log.Fatal(err)
@@ -339,6 +319,22 @@ func (s *state) ReceiveRRDPFileCallback(main string, url string, path string, da
 	}
 	f.Write(data)
 	f.Close()
+	return nil
+}
+
+func (s *state) ReceiveRRDPFileCallback(main string, url string, path string, data []byte, withdraw bool, snapshot bool, serial int64, args ...interface{}) error {
+	if len(args) > 0 {
+		rsync, ok := args[0].(string)
+		if ok && !strings.Contains(path, rsync) {
+			log.Errorf("rrdp: %s is outside directory %s", path, rsync)
+			return nil
+		}
+	}
+
+	err := s.WriteRsyncFileOnDisk(path, data, withdraw)
+	if err != nil {
+		return err
+	}
 
 	MetricSIACounts.With(
 		prometheus.Labels{
@@ -352,138 +348,182 @@ func (s *state) ReceiveRRDPFileCallback(main string, url string, path string, da
 	return nil
 }
 
-func (s *state) LoadRRDP(file string) {
+func (s *state) LoadRRDP(file string) error {
 	f, err := os.Open(file)
 	if err != nil {
-		log.Error(err)
-		return
+		return err
 	}
+	defer f.Close()
 
 	info := make(map[string]RRDPInfo)
 	dec := json.NewDecoder(f)
 	err = dec.Decode(&info)
 	if err != nil && err != io.EOF {
-		log.Error(err)
-	} else if err == nil {
-		s.RRDPInfo = info
+		return err
 	}
-	f.Close()
+	s.RRDPInfo = info
+	return nil
 }
 
-func (s *state) SaveRRDP(file string) {
+func (s *state) SaveRRDP(file string) error {
 	f, err := os.Create(file)
 	if err != nil {
-		log.Error(err)
-		return
+		return err
 	}
+	defer f.Close()
 
 	dec := json.NewEncoder(f)
-	err = dec.Encode(s.RRDPInfo)
-	if err != nil {
-		log.Error(err)
-	}
-	f.Close()
+	return dec.Encode(s.RRDPInfo)
 }
 
-func (s *state) MainRRDP() {
-	for rsync, v := range s.FinalRRDPFetch {
-		for _, vv := range v {
-			log.Infof("RRDP sync %v", vv)
+func (s *state) MainRRDP(pSpan opentracing.Span) {
+	tracer := opentracing.GlobalTracer()
+	span := tracer.StartSpan(
+		"rrdp",
+		opentracing.ChildOf(pSpan.Context()),
+	)
+	defer span.Finish()
 
-			rrdpid := vv
-			if s.RRDPMode == RRDP_MATCH_STRICT {
-				rrdpid = fmt.Sprintf("%v|%v", rsync, vv)
-			} else if s.RRDPMode == RRDP_MATCH_RSYNC {
-				newDom, _ := ExtractRsyncDomain(rsync)
-				rrdpid = fmt.Sprintf("%v|%v", newDom, vv)
-			}
+	for vv, rsync := range s.RRDPFetch {
+		rSpan := tracer.StartSpan(
+			"sync",
+			opentracing.ChildOf(span.Context()),
+		)
+		rSpan.SetTag("rrdp", vv)
+		rSpan.SetTag("rsync", rsync)
+		rSpan.SetTag("type", "rrdp")
+		log.Infof("RRDP sync %v", vv)
 
-			path := vv
-			info := s.RRDPInfo[rrdpid]
+		rrdpid := rsync
 
-			MetricSIACounts.With(
-				prometheus.Labels{
-					"address": vv,
-					"type":    "rrdp",
-				}).Set(0)
+		path := vv
+		info := s.RRDPInfo[rrdpid]
 
-			tmpStats := s.RRDPStats[vv]
-			tmpStats.URI = vv
-			tmpStats.Iteration++
-			tmpStats.Count = 0
-			s.RRDPStats[vv] = tmpStats
-			t1 := time.Now().UTC()
+		MetricSIACounts.With(
+			prometheus.Labels{
+				"address": vv,
+				"type":    "rrdp",
+			}).Set(0)
 
-			rrdp := &syncpki.RRDPSystem{
-				Callback: s.ReceiveRRDPFileCallback,
+		tmpStats := s.RRDPStats[vv]
+		tmpStats.URI = vv
+		tmpStats.Iteration++
+		tmpStats.Count = 0
+		s.RRDPStats[vv] = tmpStats
+		t1 := time.Now().UTC()
 
-				Path:    path,
-				Fetcher: s.HTTPFetcher,
+		rrdp := &syncpki.RRDPSystem{
+			Callback: s.ReceiveRRDPFileCallback,
 
-				SessionID: info.SessionID,
-				Serial:    info.Serial,
+			Path:    path,
+			Fetcher: s.HTTPFetcher,
 
-				Log: log.StandardLogger(),
-			}
-			err := rrdp.FetchRRDP(rsync)
-			t2 := time.Now().UTC()
-			if err != nil {
+			SessionID: info.SessionID,
+			Serial:    info.Serial,
+
+			Log: log.StandardLogger(),
+		}
+		err := rrdp.FetchRRDP(s.RRDPFetchDomain[vv])
+		t2 := time.Now().UTC()
+		if err != nil {
+			rSpan.SetTag("error", true)
+
+			sentry.WithScope(func(scope *sentry.Scope) {
+				if errC, ok := err.(interface{ SetURL(string, string) }); ok {
+					errC.SetURL(vv, rsync)
+				}
+				if errC, ok := err.(interface{ SetSentryScope(*sentry.Scope) }); ok {
+					errC.SetSentryScope(scope)
+				}
+				rrdp.SetSentryScope(scope)
+				scope.SetTag("Rsync", rsync)
+				scope.SetTag("RRDP", vv)
+				sentry.CaptureException(err)
+			})
+
+			if s.RRDPFailover {
 				log.Errorf("Error when processing %v (for %v): %v. Will add to rsync.", path, rsync, err)
-				s.FailoverRsync = append(s.FailoverRsync, rsync)
-
-				MetricRRDPErrors.With(
-					prometheus.Labels{
-						"address": vv,
-					}).Inc()
-
-				tmpStats = s.RRDPStats[vv]
-				tmpStats.Errors++
-				tmpStats.LastFetchError = int(time.Now().UTC().UnixNano() / 1000000000)
-				tmpStats.LastError = fmt.Sprint(err)
-				tmpStats.Duration = t2.Sub(t1).Seconds()
-				s.RRDPStats[vv] = tmpStats
-				continue
+				rSpan.LogKV("event", "rrdp failure", "type", "failover to rsync", "message", err)
+			} else {
+				log.Errorf("Error when processing %v (for %v): %v.Skipping failover to rsync.", path, rsync, err)
+				rSpan.LogKV("event", "rrdp failure", "type", "skipping failover to rsync", "message", err)
+				delete(s.RsyncFetch, rsync)
 			}
-			s.Fetcher.PathAvailable = append(s.Fetcher.PathAvailable, rsync)
-			MetricRRDPSerial.With(
+
+			MetricRRDPErrors.With(
 				prometheus.Labels{
 					"address": vv,
-				}).Set(float64(rrdp.Serial))
-			lastFetch := time.Now().UTC().UnixNano() / 1000000000
-			MetricLastFetch.With(
-				prometheus.Labels{
-					"address": vv,
-					"type":    "rrdp",
-				}).Set(float64(lastFetch))
+				}).Inc()
+
 			tmpStats = s.RRDPStats[vv]
-			tmpStats.LastFetch = int(lastFetch)
-			tmpStats.RRDPSerial = rrdp.Serial
-			tmpStats.RRDPSessionID = rrdp.SessionID
+			tmpStats.Errors++
+			tmpStats.LastFetchError = int(time.Now().UTC().UnixNano() / 1000000000)
+			tmpStats.LastError = fmt.Sprint(err)
 			tmpStats.Duration = t2.Sub(t1).Seconds()
 			s.RRDPStats[vv] = tmpStats
+			rSpan.Finish()
+			continue
+		} else {
+			log.Debugf("Success fetching %s, removing rsync %s", vv, rsync)
+			delete(s.RsyncFetch, rsync)
+		}
 
-			s.RRDPInfo[rrdpid] = RRDPInfo{
-				Rsync:     rsync,
-				Path:      path,
-				SessionID: rrdp.SessionID,
-				Serial:    rrdp.Serial,
-			}
+		rSpan.LogKV("event", "rrdp", "type", "success", "message", "rrdp successfully fetched")
+		sentry.WithScope(func(scope *sentry.Scope) {
+			scope.SetLevel(sentry.LevelInfo)
+			scope.SetTag("Rsync", rsync)
+			scope.SetTag("RRDP", vv)
+			rrdp.SetSentryScope(scope)
+			sentry.CaptureMessage("fetched rrdp successfully")
+		})
+
+		rSpan.Finish()
+		MetricRRDPSerial.With(
+			prometheus.Labels{
+				"address": vv,
+			}).Set(float64(rrdp.Serial))
+		lastFetch := time.Now().UTC().UnixNano() / 1000000000
+		MetricLastFetch.With(
+			prometheus.Labels{
+				"address": vv,
+				"type":    "rrdp",
+			}).Set(float64(lastFetch))
+		tmpStats = s.RRDPStats[vv]
+		tmpStats.LastFetch = int(lastFetch)
+		tmpStats.RRDPSerial = rrdp.Serial
+		tmpStats.RRDPSessionID = rrdp.SessionID
+		tmpStats.Duration = t2.Sub(t1).Seconds()
+		s.RRDPStats[vv] = tmpStats
+
+		s.RRDPInfo[rrdpid] = RRDPInfo{
+			Rsync:     rsync,
+			Path:      path,
+			SessionID: rrdp.SessionID,
+			Serial:    rrdp.Serial,
 		}
 	}
 }
 
-func (s *state) MainRsync() {
+func (s *state) MainRsync(pSpan opentracing.Span) {
+	tracer := opentracing.GlobalTracer()
+	span := tracer.StartSpan(
+		"rsync",
+		opentracing.ChildOf(pSpan.Context()),
+	)
+	defer span.Finish()
+
 	rsync := syncpki.RsyncSystem{
 		Log: log.StandardLogger(),
 	}
 
-	rsyncList := make([]string, 0)
-	for v, _ := range s.FinalRsyncFetch {
-		rsyncList = append(rsyncList, v)
-	}
-	rsyncList = append(rsyncList, s.FailoverRsync...)
+	for v, _ := range s.RsyncFetch {
+		rSpan := tracer.StartSpan(
+			"sync",
+			opentracing.ChildOf(span.Context()),
+		)
+		rSpan.SetTag("rsync", v)
+		rSpan.SetTag("type", "rsync")
 
-	for _, v := range rsyncList {
 		log.Infof("Rsync sync %v", v)
 		downloadPath, err := syncpki.GetDownloadPath(v, true)
 		if err != nil {
@@ -502,7 +542,20 @@ func (s *state) MainRsync() {
 		files, err := rsync.RunRsync(ctxRsync, v, s.RsyncBin, path)
 		t2 := time.Now().UTC()
 		if err != nil {
-			log.Error(err)
+			rSpan.SetTag("error", true)
+			rSpan.LogKV("event", "rsync failure", "message", err)
+			log.Errorf("Error when processing %v (for %v): %v. Will add to rsync.", path, rsync, err)
+			sentry.WithScope(func(scope *sentry.Scope) {
+				if errC, ok := err.(interface{ SetRsync(string) }); ok {
+					errC.SetRsync(v)
+				}
+				if errC, ok := err.(interface{ SetSentryScope(*sentry.Scope) }); ok {
+					errC.SetSentryScope(scope)
+				}
+				scope.SetTag("Rsync", v)
+				sentry.CaptureException(err)
+			})
+
 			MetricRsyncErrors.With(
 				prometheus.Labels{
 					"address": v,
@@ -513,13 +566,22 @@ func (s *state) MainRsync() {
 			tmpStats.LastFetchError = int(time.Now().UTC().UnixNano() / 1000000000)
 			tmpStats.LastError = fmt.Sprint(err)
 			s.RsyncStats[v] = tmpStats
+		} else {
+			rSpan.LogKV("event", "rsync", "type", "success", "message", "rsync successfully fetched")
+			sentry.WithScope(func(scope *sentry.Scope) {
+				scope.SetLevel(sentry.LevelInfo)
+				scope.SetTag("Rsync", v)
+				sentry.CaptureMessage("fetched rsync successfully")
+			})
 		}
 		cancelRsync()
 		var countFiles int
-		s.Fetcher.PathAvailable = append(s.Fetcher.PathAvailable, v)
 		if files != nil {
 			countFiles = len(files)
 		}
+
+		rSpan.Finish()
+
 		MetricSIACounts.With(
 			prometheus.Labels{
 				"address": v,
@@ -572,29 +634,226 @@ func FilterDuplicates(roalist []prefixfile.ROAJson) []prefixfile.ROAJson {
 	return roalistNodup
 }
 
-func (s *state) MainValidation() {
+func setJaegerError(l []interface{}, err error) []interface{} {
+	return append(l, "error", true, "message", err)
+}
+
+// Fetches RFC8630-type TAL
+func (s *state) MainTAL(pSpan opentracing.Span) {
+	tracer := opentracing.GlobalTracer()
+	span := tracer.StartSpan(
+		"tal",
+		opentracing.ChildOf(pSpan.Context()),
+	)
+
+	for path, tal := range s.TalsFetch {
+		tSpan := tracer.StartSpan(
+			"tal-fetch",
+			opentracing.ChildOf(span.Context()),
+		)
+		tSpan.SetTag("tal", path)
+
+		// Try the multiple URLs a TAL can be hosted on
+		var success bool
+		var successUrl string
+
+		sHub := sentry.CurrentHub().Clone()
+
+		for _, uri := range tal.URI {
+			if strings.HasPrefix(uri, "http://") || strings.HasPrefix(uri, "https://") {
+
+				tfSpan := tracer.StartSpan(
+					"tal-fetch-uri",
+					opentracing.ChildOf(tSpan.Context()),
+				)
+				tfSpan.SetTag("uri", uri)
+				//tLogs := []interface{}{"event", "fetch tal", "uri", uri}
+
+				sHub.ConfigureScope(func(scope *sentry.Scope) {
+					scope.SetTag("tal.uri", uri)
+					scope.SetTag("tal.path", path)
+				})
+
+				req, err := http.NewRequest("GET", uri, nil)
+				if err != nil {
+					tfSpan.SetTag("error", true)
+					tfSpan.SetTag("message", err)
+					tfSpan.Finish()
+					log.Errorf("error while trying to fetch: %s: %v", uri, err)
+					continue
+				}
+				req.Header.Set("User-Agent", s.HTTPFetcher.UserAgent)
+
+				sHub.ConfigureScope(func(scope *sentry.Scope) {
+					scope.SetRequest(req)
+				})
+
+				sbc := &sentry.Breadcrumb{
+					Message:  fmt.Sprintf("GET | %s", uri),
+					Category: "http",
+				}
+
+				// maybe add a limit in the client? To avoid downloading huge files (that wouldn't be certs)
+				resp, err := s.HTTPFetcher.Client.Do(req)
+				if err != nil {
+					tfSpan.SetTag("error", true)
+					tfSpan.SetTag("message", err)
+					//tSpan.LogKV(setJaegerError(tLogs, err)...)
+					tfSpan.Finish()
+
+					sbc.Level = sentry.LevelError
+					sHub.AddBreadcrumb(sbc, nil)
+					log.Errorf("error while trying to fetch: %s: %v", uri, err)
+					sHub.CaptureException(err)
+					continue
+				}
+
+				if resp.StatusCode != 200 {
+					msg := fmt.Sprintf("http server replied: %s", resp.Status)
+
+					tfSpan.SetTag("error", true)
+					tfSpan.SetTag("message", msg)
+					tfSpan.Finish()
+
+					sHub.ConfigureScope(func(scope *sentry.Scope) {
+						scope.SetLevel(sentry.LevelError)
+					})
+					sbc.Level = sentry.LevelError
+					sHub.AddBreadcrumb(sbc, nil)
+
+					log.Errorf("http server replied: %s while trying to fetch %s", resp.Status, uri)
+					sHub.CaptureMessage(msg)
+					continue
+				}
+
+				sHub.AddBreadcrumb(sbc, nil)
+
+				// check body / status code
+				data, err := ioutil.ReadAll(resp.Body)
+				tfSpan.LogKV("size", len(data))
+				if err != nil {
+					tfSpan.SetTag("error", true)
+					tfSpan.SetTag("message", err)
+					tfSpan.Finish()
+
+					log.Errorf("error while trying to fetch: %s: %v", uri, err)
+					sHub.CaptureException(err)
+					continue
+				}
+
+				// Plan option to store everything in memory
+				err = s.WriteRsyncFileOnDisk(tal.GetRsyncURI(), data, false)
+				if err != nil {
+					tfSpan.SetTag("error", true)
+					tfSpan.SetTag("message", err)
+					tfSpan.Finish()
+
+					log.Errorf("error while trying to fetch: %s: %v", uri, err)
+					sHub.CaptureException(err)
+					continue
+				}
+
+				//tSpan.LogKV(append(tLogs, "success", true)...)
+				tfSpan.Finish()
+
+				sHub.WithScope(func(scope *sentry.Scope) {
+					scope.SetLevel(sentry.LevelInfo)
+					sHub.CaptureMessage("fetched http tal cert successfully")
+				})
+
+				success = true
+				successUrl = uri
+				break
+
+			}
+		}
+
+		// Fail over to rsync
+		if !success && s.RRDPFailover && tal.HasRsync() {
+			rsync := tal.GetRsyncURI()
+			log.Infof("Root certificate for %s will be downloaded using rsync: %s", path, rsync)
+			s.RsyncFetch[rsync] = ""
+			tSpan.SetTag("failover-rsync", true)
+		} else if success {
+			log.Infof("Successfully downloaded root certificate for %s at %s", path, successUrl)
+		} else {
+			log.Errorf("Could not download root certificate for %s", path)
+			tSpan.SetTag("error", true)
+		}
+
+		tSpan.Finish()
+	}
+
+	defer span.Finish()
+}
+
+func (s *state) MainValidation(pSpan opentracing.Span) {
+	ia := make([][]SIA, len(s.Tals))
+	for i := 0; i < len(ia); i++ {
+		ia[i] = make([]SIA, 0)
+	}
+	iatmp := make(map[string]*SIA)
+
+	tracer := opentracing.GlobalTracer()
+	span := tracer.StartSpan(
+		"validation",
+		opentracing.ChildOf(pSpan.Context()),
+	)
+	defer span.Finish()
+
 	manager := make([]*pki.SimpleManager, len(s.Tals))
 	for i, tal := range s.Tals {
-		validator := pki.NewValidator()
+		tSpan := tracer.StartSpan(
+			"explore",
+			opentracing.ChildOf(span.Context()),
+		)
+		tSpan.SetTag("tal", tal.Path)
 
-		manager[i] = pki.NewSimpleManager()
+		validator := pki.NewValidator()
+		validator.DecoderConfig.ValidateStrict = s.StrictCms
+
+		sm := pki.NewSimpleManager()
+		manager[i] = sm
+		manager[i].ReportErrors = true
 		manager[i].Validator = validator
 		manager[i].FileSeeker = s.Fetcher
 		manager[i].Log = s
+		manager[i].StrictHash = s.StrictHash
+		manager[i].StrictManifests = s.StrictManifests
 
+		go func(sm *pki.SimpleManager) {
+			for err := range sm.Errors {
+				tSpan.SetTag("error", true)
+				tSpan.LogKV("event", "resource issue", "type", "skipping resource", "message", err)
+				//log.Errorf("Error when processing %v (for %v): %v.", path, rsync, err)
+				log.Error(err)
+				sentry.WithScope(func(scope *sentry.Scope) {
+					if errC, ok := err.(interface{ SetSentryScope(*sentry.Scope) }); ok {
+						errC.SetSentryScope(scope)
+					}
+					scope.SetTag("TrustAnchor", tal.Path)
+					sentry.CaptureException(err)
+				})
+			}
+
+			//log.Warn("Closed errors")
+		}(sm)
 		manager[i].AddInitial([]*pki.PKIFile{tal})
 		s.CountExplore = manager[i].Explore(!s.UseManifest, false)
 
 		// Insertion of SIAs in db to allow rsync to update the repos
 		var count int
 		for _, obj := range manager[i].Validator.TALs {
-			tal := obj.Resource.(*librpki.RPKI_TAL)
-			s.RsyncFetch[tal.URI] = time.Now().UTC()
+			tal := obj.Resource.(*librpki.RPKITAL)
+			//s.RsyncFetch[tal.GetURI()] = time.Now().UTC()
+			if !obj.CertTALValid {
+				s.TalsFetch[obj.File.Path] = tal
+			}
 			count++
 		}
 		for _, obj := range manager[i].Validator.ValidObjects {
 			if obj.Type == pki.TYPE_CER {
-				cer := obj.Resource.(*librpki.RPKI_Certificate)
+				cer := obj.Resource.(*librpki.RPKICertificate)
 				var RsyncGN string
 				var RRDPGN string
 				var hasRRDP bool
@@ -602,25 +861,55 @@ func (s *state) MainValidation() {
 					gn := string(sia.GeneralName)
 					if sia.AccessMethod.Equal(CertRepository) {
 						RsyncGN = gn
-						s.RsyncFetch[gn] = time.Now().UTC()
 					} else if sia.AccessMethod.Equal(CertRRDP) {
 						hasRRDP = true
 						RRDPGN = gn
 					}
 				}
-
-				if hasRRDP {
-					if _, ok := s.RRDPFetch[RRDPGN]; !ok {
-						s.RRDPFetch[RRDPGN] = make([]string, 0)
-					}
-					s.RRDPFetch[RRDPGN] = append(s.RRDPFetch[RRDPGN], RsyncGN)
+				gnExtracted, gnExtractedDomain, err := syncpki.ExtractRsyncDomainModule(RsyncGN)
+				if err != nil {
+					log.Errorf("Could not add cert rsync %s due to %v", RsyncGN, err)
+					continue
 				}
 
+				if hasRRDP {
+					prev, ok := s.RRDPFetchDomain[RRDPGN]
+					if ok && prev != gnExtractedDomain {
+						log.Errorf("rrdp %s tries to override %s with %s", RRDPGN, prev, gnExtractedDomain)
+						continue
+					}
+					s.RRDPFetchDomain[RRDPGN] = gnExtractedDomain
+					s.RRDPFetch[RRDPGN] = gnExtracted
+				}
+				s.RsyncFetch[gnExtracted] = RRDPGN
+				s.CurrentRepos[gnExtracted] = time.Now().UTC()
 				count++
+
+				// map the rrdp and rsync by TAL for info page
+				iaId, ok := iatmp[gnExtracted]
+				if !ok {
+					iaIdTmp := SIA{
+						gnExtracted,
+						RRDPGN,
+					}
+					ia[i] = append(ia[i], iaIdTmp)
+					iaId = &(ia[i][len(ia[i])-1])
+					iatmp[gnExtracted] = iaId
+				}
+				iaId.Rsync = gnExtracted
+				iaId.RRDP = RRDPGN
+
 			}
 		}
+		sm.Close()
+		tSpan.LogKV("count-valid", count, "count-total", s.CountExplore)
+		tSpan.Finish()
 	}
 
+	s.InfoAuthoritiesLock.Lock()
+	s.InfoAuthorities = ia
+	s.InfoAuthoritiesLock.Unlock()
+
 	// Generating ROAs list
 	roalist := &prefixfile.ROAList{
 		Data: make([]prefixfile.ROAJson, 0),
@@ -628,6 +917,11 @@ func (s *state) MainValidation() {
 	var counts int
 	s.ROAsTALsCount = make([]ROAsTAL, 0)
 	for i, tal := range s.Tals {
+		eSpan := tracer.StartSpan(
+			"extract",
+			opentracing.ChildOf(span.Context()),
+		)
+		eSpan.SetTag("tal", tal.Path)
 		talname := tal.Path
 		if len(s.TalNames) == len(s.Tals) {
 			talname = s.TalNames[i]
@@ -635,7 +929,7 @@ func (s *state) MainValidation() {
 
 		var counttal int
 		for _, obj := range manager[i].Validator.ValidROA {
-			roa := obj.Resource.(*librpki.RPKI_ROA)
+			roa := obj.Resource.(*librpki.RPKIROA)
 
 			for _, entry := range roa.Valids {
 				oroa := prefixfile.ROAJson{
@@ -649,6 +943,8 @@ func (s *state) MainValidation() {
 				counttal++
 			}
 		}
+		eSpan.Finish()
+
 		s.ROAsTALsCount = append(s.ROAsTALsCount, ROAsTAL{TA: talname, Count: counttal})
 		MetricROAsCount.With(
 			prometheus.Labels{
@@ -666,19 +962,29 @@ func (s *state) MainValidation() {
 
 	roalist.Data = FilterDuplicates(roalist.Data)
 	if s.Sign {
+		sSpan := tracer.StartSpan(
+			"sign",
+			opentracing.ChildOf(span.Context()),
+		)
+
 		signdate, sign, err := roalist.Sign(s.Key)
 		if err != nil {
 			log.Error(err)
+			sentry.CaptureException(err)
 		}
 		roalist.Metadata.Signature = sign
 		roalist.Metadata.SignatureDate = signdate
+
+		sSpan.Finish()
 	}
 
+	s.ROAListLock.Lock()
 	s.ROAList = roalist
+	s.ROAListLock.Unlock()
 }
 
 func (s *state) ServeROAs(w http.ResponseWriter, r *http.Request) {
-	if s.Stable || !s.WaitStable {
+	if s.Stable || !s.WaitStable || s.HasPreviousStable {
 
 		upTo := s.LastComputed.Add(s.Validity)
 		maxAge := int(upTo.Sub(time.Now()).Seconds())
@@ -689,7 +995,9 @@ func (s *state) ServeROAs(w http.ResponseWriter, r *http.Request) {
 			w.Header().Set("Cache-Control", fmt.Sprintf("max-age=%v", maxAge))
 		}
 
+		s.ROAListLock.RLock()
 		tmp := s.ROAList
+		s.ROAListLock.RUnlock()
 
 		etag := sha256.New()
 		etag.Write([]byte(fmt.Sprintf("%v/%v", tmp.Metadata.Generated, tmp.Metadata.Counts)))
@@ -712,6 +1020,15 @@ func (s *state) ServeROAs(w http.ResponseWriter, r *http.Request) {
 	}
 }
 
+func (s *state) ServeHealth(w http.ResponseWriter, r *http.Request) {
+	if s.Stable || s.HasPreviousStable {
+		w.WriteHeader(http.StatusOK)
+		return
+	}
+	w.WriteHeader(http.StatusServiceUnavailable)
+	w.Write([]byte("Not ready yet"))
+}
+
 type SIA struct {
 	Rsync string `json:"rsync"`
 	RRDP  string `json:"rrdp,omitempty"`
@@ -722,78 +1039,63 @@ type ROAsTAL struct {
 	Count int    `json:"count,omitempty"`
 }
 
+type InfoAuthorities struct {
+	TA  string `json:"name"`
+	Sia []SIA  `json:"sia"`
+}
+
 type InfoResult struct {
-	Stable             bool      `json:"stable"`
-	TALs               []string  `json:"tals"`
-	SIAs               []SIA     `json:"sia"`
-	Rsync              []Stats   `json:"sias-rsync,omitempty"`
-	RRDP               []Stats   `json:"sias-rrdp,omitempty"`
-	Iteration          int       `json:"iteration"`
-	LastValidation     int       `json:"validation-last"`
-	ValidationDuration float64   `json:"validation-duration"`
-	ValidationMessages []string  `json:"validation-messages"`
-	ExploredFiles      int       `json:"validation-explored"`
-	ROAsTALs           []ROAsTAL `json:"roas-tal-count"`
-	ROACount           int       `json:"roas-count"`
+	Stable             bool              `json:"stable"`
+	TAs                []InfoAuthorities `json:"tas"`
+	Iteration          int               `json:"iteration"`
+	LastValidation     int               `json:"validation-last"`
+	ValidationDuration float64           `json:"validation-duration"`
+	ROAsTALs           []ROAsTAL         `json:"roas-tal-count"`
+	ROACount           int               `json:"roas-count"`
 }
 
 func (s *state) ServeInfo(w http.ResponseWriter, r *http.Request) {
 	w.Header().Set("Content-Type", "application/json")
-	tmproa := s.ROAList
 
-	sia := make([]SIA, 0)
-	tmprsyncfetch := s.FinalRsyncFetch
-	tmprrdpfetch := s.FinalRRDPFetch
-	for k, _ := range tmprsyncfetch {
-		sia = append(sia, SIA{
-			Rsync: k,
-		})
-	}
-	for k, v := range tmprrdpfetch {
-		for _, vv := range v {
-			sia = append(sia, SIA{
-				Rsync: k,
-				RRDP:  vv,
-			})
+	s.InfoAuthoritiesLock.RLock()
+	ia := s.InfoAuthorities
+	s.InfoAuthoritiesLock.RUnlock()
+
+	ias := make([]InfoAuthorities, 0)
+	for i, tal := range s.Tals {
+
+		if len(ia) <= i {
+			break
+		}
+		if ia[i] == nil {
+			continue
 		}
-	}
-	tmprsync := s.RsyncStats
-	tmprrdp := s.RRDPStats
-	tmprsyncstats := make([]Stats, 0)
-	tmprrdpstats := make([]Stats, 0)
-	for _, v := range tmprsync {
-		tmprsyncstats = append(tmprsyncstats, v)
-	}
-	for _, v := range tmprrdp {
-		tmprrdpstats = append(tmprrdpstats, v)
-	}
-	vm := s.ValidationMessages
 
-	tals := make([]string, 0)
-	tmptals := s.Tals
-	for _, v := range tmptals {
-		tals = append(tals, v.Path)
+		talname := tal.Path
+		if len(s.TalNames) == len(s.Tals) {
+			talname = s.TalNames[i]
+		}
+
+		ias = append(ias, InfoAuthorities{
+			TA:  talname,
+			Sia: ia[i],
+		})
 	}
 
 	ir := InfoResult{
-		TALs:               tals,
-		Stable:             s.Stable,
-		SIAs:               sia,
-		ROACount:           len(tmproa.Data),
+		TAs:                ias,
+		ROACount:           len(s.ROAList.Data),
 		ROAsTALs:           s.ROAsTALsCount,
-		Rsync:              tmprsyncstats,
-		RRDP:               tmprrdpstats,
+		Stable:             s.Stable,
 		LastValidation:     int(s.LastComputed.UnixNano() / 1000000),
-		ExploredFiles:      s.CountExplore,
 		ValidationDuration: s.ValidationDuration.Seconds(),
 		Iteration:          s.Iteration,
-		ValidationMessages: vm,
 	}
 	enc := json.NewEncoder(w)
 	enc.Encode(ir)
 }
 
-func (s *state) Serve(addr string, path string, metricsPath string, infoPath string, corsOrigin string, corsCreds bool) {
+func (s *state) Serve(addr string, path string, metricsPath string, infoPath string, healthPath string, corsOrigin string, corsCreds bool) {
 	// Note(Erica): fix https://github.com/cloudflare/cfrpki/issues/8
 	fullPath := path
 	if len(path) > 0 && string(path[0]) != "/" {
@@ -801,23 +1103,32 @@ func (s *state) Serve(addr string, path string, metricsPath string, infoPath str
 	}
 	log.Infof("Serving HTTP on %v%v", addr, fullPath)
 
-	r := mux.NewRouter()
+	r := http.NewServeMux()
 
 	r.HandleFunc(fullPath, s.ServeROAs)
 	r.HandleFunc(infoPath, s.ServeInfo)
+	r.HandleFunc(healthPath, s.ServeHealth)
 	r.Handle(metricsPath, promhttp.Handler())
 
+	if s.Pprof {
+		r.HandleFunc("/debug/pprof/cmdline", pprof.Cmdline)
+		r.HandleFunc("/debug/pprof/profile", pprof.Profile)
+		r.HandleFunc("/debug/pprof/symbol", pprof.Symbol)
+		r.HandleFunc("/debug/pprof/trace", pprof.Trace)
+		r.HandleFunc("/debug/pprof/", pprof.Index)
+	}
+
 	corsReq := cors.New(cors.Options{
 		AllowedOrigins:   strings.Split(corsOrigin, ","),
 		AllowedMethods:   []string{"GET", "POST", "OPTIONS"},
 		AllowCredentials: corsCreds,
 	}).Handler(r)
 
-	http.Handle("/", corsReq)
-	log.Fatal(http.ListenAndServe(addr, nil))
+	log.Fatal(http.ListenAndServe(addr, corsReq))
 }
 
 func init() {
+
 	prometheus.MustRegister(MetricSIACounts)
 	prometheus.MustRegister(MetricRsyncErrors)
 	prometheus.MustRegister(MetricRRDPErrors)
@@ -841,7 +1152,35 @@ func main() {
 
 	lvl, _ := log.ParseLevel(*LogLevel)
 	log.SetLevel(lvl)
-	log.Infof("Validator started")
+
+	sentryDsn := *SentryDSN
+	if sentryDsn == "" {
+		sentryDsn = os.Getenv("SENTRY_DSN")
+	}
+	if sentryDsn != "" {
+		err := sentry.Init(sentry.ClientOptions{
+			Dsn: sentryDsn,
+		})
+		if err != nil {
+			log.Fatalf("failed initializing sentry: %s", err)
+		}
+		defer sentry.Flush(2 * time.Second)
+	}
+
+	log.Info("Validator started")
+
+	if *Tracer {
+		cfg, err := jcfg.FromEnv()
+		if err != nil {
+			log.Fatal(err)
+		}
+		tracer, closer, err := cfg.NewTracer()
+		if err != nil {
+			log.Fatal(err)
+		}
+		defer closer.Close()
+		opentracing.SetGlobalTracer(tracer)
+	}
 
 	mainRefresh, _ := time.ParseDuration(*Refresh)
 
@@ -857,9 +1196,15 @@ func main() {
 	timeoutDur, _ := time.ParseDuration(*RsyncTimeout)
 	timeValidity, _ := time.ParseDuration(*Validity)
 
+	err := os.MkdirAll(*Basepath, os.ModePerm)
+	if err != nil {
+		log.Fatal(err)
+	}
+
 	s := &state{
 		Basepath:     *Basepath,
 		Tals:         tals,
+		TalsFetch:    make(map[string]*librpki.RPKITAL),
 		TalNames:     talNames,
 		UseManifest:  *UseManifest,
 		RsyncTimeout: timeoutDur,
@@ -871,34 +1216,44 @@ func main() {
 
 		EnableCache: *CacheHeader,
 
-		Mode:     *Mode,
-		RRDPMode: *RRDPMode,
+		Mode:         *Mode,
+		RRDPFailover: *RRDPFailover,
 
-		RsyncFetch:      make(map[string]time.Time),
-		RRDPFetch:       make(map[string][]string),
-		FinalRsyncFetch: make(map[string]bool),
-		FinalRRDPFetch:  make(map[string][]string),
-		RRDPInfo:        make(map[string]RRDPInfo),
-		FailoverRsync:   make([]string, 0),
+		RRDPInfo: make(map[string]RRDPInfo),
 
-		Fetcher: &syncpki.LocalFetch{
-			MapDirectory: map[string]string{
+		PrevRepos:    make(map[string]time.Time),
+		CurrentRepos: make(map[string]time.Time),
+
+		RsyncFetch:      make(map[string]string),
+		RRDPFetch:       make(map[string]string),
+		RRDPFetchDomain: make(map[string]string),
+
+		Fetcher: syncpki.NewLocalFetch(
+			map[string]string{
 				"rsync://": *Basepath,
 			},
-			Log:           log.StandardLogger(),
-			PathAvailable: make([]string, 0),
-		},
+			log.StandardLogger()),
 		HTTPFetcher: &syncpki.HTTPFetcher{
-			UserAgent: "Cloudflare-RPKI-RRDP/1.0 (+https://rpki.cloudflare.com)",
+			UserAgent: *UserAgent,
 			Client:    &http.Client{},
 		},
 		ROAList: &prefixfile.ROAList{
 			Data: make([]prefixfile.ROAJson, 0),
 		},
+		ROAListLock: &sync.RWMutex{},
 
 		RsyncStats:    make(map[string]Stats),
 		RRDPStats:     make(map[string]Stats),
 		ROAsTALsCount: make([]ROAsTAL, 0),
+
+		InfoAuthorities:     make([][]SIA, 0),
+		InfoAuthoritiesLock: &sync.RWMutex{},
+
+		Pprof: *Pprof,
+
+		StrictHash:      *StrictHash,
+		StrictManifests: *StrictManifests,
+		StrictCms:       *StrictCms,
 	}
 
 	if *Sign {
@@ -919,24 +1274,43 @@ func main() {
 	}
 
 	if *Mode == "server" {
-		go s.Serve(*Addr, *Output, *MetricsPath, *InfoPath, *CorsOrigins, *CorsCreds)
+		go s.Serve(*Addr, *Output, *MetricsPath, *InfoPath, *HealthPath, *CorsOrigins, *CorsCreds)
 	} else if *Mode != "oneoff" {
 		log.Fatalf("Mode %v is not specified. Choose either server or oneoff", *Mode)
 	}
+	tracer := opentracing.GlobalTracer()
 
+	var spanActive bool
+	var pSpan opentracing.Span
+	var iterationsUntilStable int
 	for {
+		if !spanActive {
+			pSpan = tracer.StartSpan("multoperation")
+			spanActive = true
+			iterationsUntilStable = 0
+		}
+
+		span := tracer.StartSpan("operation", opentracing.ChildOf(pSpan.Context()))
+
 		s.Iteration++
-		s.FailoverRsync = make([]string, 0)
-		s.Fetcher.PathAvailable = make([]string, 0)
+		iterationsUntilStable++
+		span.SetTag("iteration", s.Iteration)
+
 		if *RRDP {
 			t1 := time.Now().UTC()
 			// RRDP
 			if *RRDPFile != "" {
-				s.LoadRRDP(*RRDPFile)
+				err = s.LoadRRDP(*RRDPFile)
+				if err != nil {
+					sentry.CaptureException(err)
+				}
 			}
-			s.MainRRDP()
+			s.MainRRDP(span)
 			if *RRDPFile != "" {
 				s.SaveRRDP(*RRDPFile)
+				if err != nil {
+					sentry.CaptureException(err)
+				}
 			}
 
 			t2 := time.Now().UTC()
@@ -949,10 +1323,23 @@ func main() {
 
 		t1 := time.Now().UTC()
 
-		// Rsync
-		s.MainRsync()
+		// HTTPs TAL
+		s.MainTAL(span)
+		s.TalsFetch = make(map[string]*librpki.RPKITAL) // clear decoded TAL for next iteration
 
 		t2 := time.Now().UTC()
+		MetricOperationTime.With(
+			prometheus.Labels{
+				"type": "tal",
+			}).
+			Observe(float64(t2.Sub(t1).Seconds()))
+
+		t1 = time.Now().UTC()
+
+		// Rsync
+		s.MainRsync(span)
+
+		t2 = time.Now().UTC()
 		MetricOperationTime.With(
 			prometheus.Labels{
 				"type": "rsync",
@@ -963,7 +1350,7 @@ func main() {
 		t1 = time.Now().UTC()
 
 		// Validation
-		s.MainValidation()
+		s.MainValidation(span)
 
 		t2 = time.Now().UTC()
 		s.ValidationDuration = t2.Sub(t1)
@@ -977,7 +1364,8 @@ func main() {
 		t1 = time.Now().UTC()
 
 		// Reduce
-		s.Stable = !s.MainReduce()
+		s.Stable = !s.MainReduce() && s.Iteration > 1
+		s.HasPreviousStable = s.Stable
 
 		t2 = time.Now().UTC()
 		MetricOperationTime.With(
@@ -1002,6 +1390,9 @@ func main() {
 
 		}
 
+		span.SetTag("stable", s.Stable)
+		span.Finish()
+
 		if *Mode == "oneoff" && s.Stable {
 			log.Info("Stable, terminating")
 			break
@@ -1011,6 +1402,10 @@ func main() {
 			MetricLastStableValidation.Set(float64(s.LastComputed.UnixNano() / 1000000000))
 			MetricState.Set(float64(1))
 
+			pSpan.SetTag("iterations", iterationsUntilStable)
+			pSpan.Finish()
+			spanActive = false
+
 			log.Infof("Stable state. Revalidating in %v", mainRefresh)
 			<-time.After(mainRefresh)
 			s.Stable = false
@@ -1019,5 +1414,6 @@ func main() {
 
 			log.Info("Still exploring. Revalidating now")
 		}
+
 	}
 }
