@@ -1,10 +1,10 @@
 package pki
 
 import (
+	"bytes"
 	"crypto/x509"
 	"crypto/x509/pkix"
 	"encoding/asn1"
-	"encoding/hex"
 	"errors"
 	"fmt"
 	"github.com/cloudflare/cfrpki/validator/lib"
@@ -26,6 +26,18 @@ const (
 var (
 	CARepository = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 48, 5}
 	Manifest     = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 48, 10}
+
+	TypeToName = map[int]string{
+		TYPE_UNKNOWN: "unknown",
+		TYPE_CER:     "certificate",
+		TYPE_MFT:     "manifest",
+		TYPE_ROA:     "roa",
+		TYPE_CRL:     "crl",
+		TYPE_ROACER:  "roa-ee",
+		TYPE_MFTCER:  "manifest-ee",
+		TYPE_CAREPO:  "ca-repo",
+		TYPE_TAL:     "tal",
+	}
 )
 
 type Resource struct {
@@ -34,24 +46,27 @@ type Resource struct {
 	File     *PKIFile
 	Resource interface{}
 	Childs   []*Resource
+
+	CertTALValid bool // currently used for TALs: indicates the child is valid and does not need to be fetched again
 }
 
 func (res *Resource) GetIdentifier() (bool, []byte) {
 	switch res := res.Resource.(type) {
-	case *librpki.RPKI_Certificate:
+	case *librpki.RPKICertificate:
 		return true, res.Certificate.SubjectKeyId
-	case *librpki.RPKI_ROA:
+	case *librpki.RPKIROA:
 		return true, res.Certificate.Certificate.SubjectKeyId
-	case *librpki.RPKI_Manifest:
+	case *librpki.RPKIManifest:
 		return true, res.Certificate.Certificate.SubjectKeyId
 	}
 	return false, nil
 }
 
 type SeekFile struct {
-	Repo string
-	File string
-	Data []byte
+	Repo   string
+	File   string
+	Data   []byte
+	Sha256 []byte
 }
 
 type FileSeeker interface {
@@ -68,22 +83,49 @@ type Log interface {
 
 type SimpleManager struct {
 	PathOfResource  map[*Resource]*PKIFile
+	ResourceOfPath  map[*PKIFile]*Resource
 	ToExplore       []*PKIFile
 	FileSeeker      FileSeeker
 	Validator       *Validator
 	Explored        map[string]bool
 	ToExploreUnique map[string]bool
 	Log             Log
+
+	ReportErrors bool
+	Errors       chan error
+
+	StrictManifests bool
+	StrictHash      bool
 }
 
 func NewSimpleManager() *SimpleManager {
 	return &SimpleManager{
 		PathOfResource:  make(map[*Resource]*PKIFile),
+		ResourceOfPath:  make(map[*PKIFile]*Resource),
 		Explored:        make(map[string]bool),
 		ToExploreUnique: make(map[string]bool),
+		Errors:          make(chan error, 50),
+		StrictManifests: true,
+		StrictHash:      true,
 	}
 }
 
+func (sm *SimpleManager) Close() {
+	close(sm.Errors)
+}
+
+func (sm *SimpleManager) reportError(err error) {
+	if sm.ReportErrors {
+		sm.Errors <- err
+	}
+}
+func (sm *SimpleManager) reportErrorFile(err error, file *PKIFile, seek *SeekFile) {
+	if errC, ok := err.(interface{ AddFileErrorInfo(*PKIFile, *SeekFile) }); file != nil && ok {
+		errC.AddFileErrorInfo(file, seek)
+	}
+	sm.reportError(err)
+}
+
 func (sm *SimpleManager) PutFiles(fileList []*PKIFile) {
 	for _, file := range fileList {
 		path := file.ComputePath()
@@ -121,6 +163,9 @@ func (sm *SimpleManager) GetNextFile(curExplore *PKIFile) (*SeekFile, error) {
 
 	if sm.FileSeeker != nil {
 		data, err := sm.FileSeeker.GetFile(curExplore)
+		if err != nil {
+			err = NewFileError(err)
+		}
 		return data, err
 	}
 	return nil, errors.New("No interface to fetch file, check FileSeeker")
@@ -147,6 +192,9 @@ type Validator struct {
 	ValidObjects map[string]*Resource
 	Objects      map[string]*Resource
 
+	// Key by path
+	ObjectsPath map[string]*Resource
+
 	CertsSerial map[string]*Resource
 	Revoked     map[string]bool
 
@@ -162,6 +210,8 @@ type Validator struct {
 	ValidManifest map[string]*Resource // Make sure EE certificates are unique for a ROA
 	Manifest      map[string]*Resource
 
+	DecoderConfig *librpki.DecoderConfig
+
 	Time time.Time
 }
 
@@ -172,6 +222,8 @@ func NewValidator() *Validator {
 		ValidObjects: make(map[string]*Resource),
 		Objects:      make(map[string]*Resource),
 
+		ObjectsPath: make(map[string]*Resource),
+
 		CertsSerial: make(map[string]*Resource),
 		Revoked:     make(map[string]bool),
 
@@ -184,6 +236,8 @@ func NewValidator() *Validator {
 		ValidManifest: make(map[string]*Resource),
 		Manifest:      make(map[string]*Resource),
 
+		DecoderConfig: librpki.DefaultDecoderConfig,
+
 		Time: time.Now().UTC(),
 	}
 }
@@ -194,6 +248,8 @@ type PKIFile struct {
 	Path   string
 	Type   int
 	Trust  bool
+
+	ManifestHash []byte
 }
 
 func (f *PKIFile) ComputePath() string {
@@ -226,7 +282,7 @@ func (v *Validator) AddResource(pkifile *PKIFile, data []byte) (bool, []*PKIFile
 		}
 		pathCert, res, err := v.AddTAL(tal)
 		if res == nil {
-			return true, pathCert, res, errors.New(fmt.Sprintf("Resource is empty: %v", err))
+			return true, pathCert, res, errors.New("Resource is empty")
 		}
 		res.File = pkifile
 		for _, pc := range pathCert {
@@ -241,47 +297,55 @@ func (v *Validator) AddResource(pkifile *PKIFile, data []byte) (bool, []*PKIFile
 		if pkifile != nil && pkifile.Parent != nil && pkifile.Parent.Type == TYPE_TAL {
 			talComp, ok := v.TALs[pkifile.Path]
 			if ok {
-				talValidation := talComp.Resource.(*librpki.RPKI_TAL).CheckCertificate(cert.Certificate)
+				talValidation := talComp.Resource.(*librpki.RPKITAL).CheckCertificate(cert.Certificate)
 				if !talValidation {
 					return false, nil, nil, errors.New("Certificate was not validated against TAL")
 				}
+				v.TALs[pkifile.Path].CertTALValid = true // indicates that we can skip downloading
 			}
 		}
 
 		valid, pathCert, res, err := v.AddCert(cert, pkifile.Trust)
 		if res == nil {
-			return valid, pathCert, res, errors.New(fmt.Sprintf("Resource is empty: %v", err))
+			return valid, pathCert, res, fmt.Errorf("Resource is empty: %v", err)
 		}
 		res.Type = TYPE_CER
 		res.File = pkifile
 		for _, pc := range pathCert {
 			pc.Parent = pkifile
 		}
+
+		v.ObjectsPath[pkifile.Path] = res
 		return valid, pathCert, res, err
 	case TYPE_ROA:
-		roa, err := librpki.DecodeROA(data)
+		roa, err := v.DecoderConfig.DecodeROA(data)
 		if err != nil {
 			return false, nil, nil, err
 		}
 		valid, res, err := v.AddROA(pkifile, roa)
 		if res == nil {
-			return valid, nil, res, errors.New(fmt.Sprintf("Resource is empty: %v", err))
+			return valid, nil, res, fmt.Errorf("Resource is empty: %v", err)
 		}
 		res.File = pkifile
+
+		v.ObjectsPath[pkifile.Path] = res
 		return valid, nil, res, err
 	case TYPE_MFT:
-		mft, err := librpki.DecodeManifest(data)
+		mft, err := v.DecoderConfig.DecodeManifest(data)
 		if err != nil {
 			return false, nil, nil, err
 		}
 		valid, pathCert, res, err := v.AddManifest(pkifile, mft)
 		if res == nil {
-			return valid, nil, res, errors.New(fmt.Sprintf("Resource is empty: %v", err))
+			return valid, nil, res, fmt.Errorf("Resource is empty: %v", err)
 		}
 		res.File = pkifile
+		// add the parent information to invalidate the Manifest in case of an issue
 		for _, pc := range pathCert {
 			pc.Parent = pkifile
 		}
+
+		v.ObjectsPath[pkifile.Path] = res
 		return valid, pathCert, res, err
 	case TYPE_CRL:
 		// https://tools.ietf.org/html/rfc5280
@@ -290,10 +354,15 @@ func (v *Validator) AddResource(pkifile *PKIFile, data []byte) (bool, []*PKIFile
 			return false, nil, nil, err
 		}
 		valid, res, err := v.AddCRL(crl)
+		if pkifile.Parent.Parent.Path != res.Parent.File.Path {
+			return false, nil, nil, fmt.Errorf("CRL %s does not match with the parent %s", pkifile.Path, pkifile.Parent.Parent.Path)
+		}
 		if res == nil {
-			return valid, nil, res, errors.New(fmt.Sprintf("Resource is empty: %v", err))
+			return valid, nil, res, fmt.Errorf("Resource is empty: %v", err)
 		}
 		res.File = pkifile
+
+		v.ObjectsPath[pkifile.Path] = res
 		return valid, nil, res, err
 	}
 	return false, nil, nil, errors.New("Unknown file type")
@@ -333,8 +402,8 @@ func (v *Validator) InvalidateObject(keyid []byte) {
 
 }
 
-func (v *Validator) AddTAL(tal *librpki.RPKI_TAL) ([]*PKIFile, *Resource, error) {
-	uri := tal.URI
+func (v *Validator) AddTAL(tal *librpki.RPKITAL) ([]*PKIFile, *Resource, error) {
+	uri := tal.GetRsyncURI()
 	files := []*PKIFile{
 		&PKIFile{
 			Type:  TYPE_CER,
@@ -350,20 +419,22 @@ func (v *Validator) AddTAL(tal *librpki.RPKI_TAL) ([]*PKIFile, *Resource, error)
 	return files, res, nil
 }
 
-func (v *Validator) AddCert(cert *librpki.RPKI_Certificate, trust bool) (bool, []*PKIFile, *Resource, error) {
+func (v *Validator) AddCert(cert *librpki.RPKICertificate, trust bool) (bool, []*PKIFile, *Resource, error) {
 	pathCert := ExtractPathCert(cert)
 
 	ski := string(cert.Certificate.SubjectKeyId)
 	aki := string(cert.Certificate.AuthorityKeyId)
 
-	_, exists := v.Objects[ski]
+	res := ObjectToResource(cert)
+
+	conflict, exists := v.Objects[ski]
 	if exists {
-		return false, nil, nil, errors.New(fmt.Sprintf("A certificate with Subject Key Id: %v already exists", hex.EncodeToString))
+		conflictCert, _ := conflict.Resource.(*librpki.RPKICertificate)
+		return false, nil, res, NewCertificateErrorConflict(cert, conflictCert)
 	}
 
 	_, hasParentValid := v.ValidObjects[aki]
 	parent, hasParent := v.Objects[aki]
-	res := ObjectToResource(cert)
 	res.Parent = parent
 
 	var valid bool
@@ -390,13 +461,11 @@ func (v *Validator) AddCert(cert *librpki.RPKI_Certificate, trust bool) (bool, [
 	return valid, pathCert, res, err
 }
 
-func (v *Validator) ValidateCertificate(cert *librpki.RPKI_Certificate, trust bool) error {
-	ski := cert.Certificate.SubjectKeyId
-
+func (v *Validator) ValidateCertificate(cert *librpki.RPKICertificate, trust bool) error {
 	// Check time validity
 	err := cert.ValidateTime(v.Time)
 	if err != nil {
-		return errors.New(fmt.Sprintf("Could not validate certificate due to expiration date %x: %v", ski, err))
+		return NewCertificateErrorValidity(cert, err)
 	}
 
 	if trust {
@@ -407,74 +476,76 @@ func (v *Validator) ValidateCertificate(cert *librpki.RPKI_Certificate, trust bo
 	aki := cert.Certificate.AuthorityKeyId
 	parent, hasParent := v.ValidObjects[string(aki)]
 	if !hasParent {
-		return errors.New(fmt.Sprintf("Could not find parent %x for certificate %x", aki, ski))
+		return NewCertificateErrorParent(cert, nil, errors.New("missing parent"))
 	}
 
-	parentCert, ok := parent.Resource.(*librpki.RPKI_Certificate)
+	parentCert, ok := parent.Resource.(*librpki.RPKICertificate)
 	if !ok {
-		return errors.New(fmt.Sprintf("Parent %x of %x is not a RPKI Certificate", ski, aki))
+		return NewCertificateErrorParent(cert, parentCert, errors.New("parent is not a rpki certificate"))
 	}
 	err = cert.Validate(parentCert)
 	if err != nil {
-		return errors.New(fmt.Sprintf("Could not validate certificate %x against parent %x: %v", ski, aki, err))
+		return NewCertificateErrorParent(cert, parentCert, err)
 	}
 
-	// Check presence in revokation lists
+	// Check presence in revocation lists
 	_, revoked := v.Revoked[string(aki)+cert.Certificate.SerialNumber.String()]
 	if revoked {
-		return errors.New(fmt.Sprintf("Certificate was revoked by issuer %x", ski))
+		return NewCertificateErrorRevocation(cert)
 	}
 
 	// Check IPs
-	valids, invalids, checkParent := cert.ValidateIPCertificate(parentCert)
+	validIPs, invalidIPs, checkParent := cert.ValidateIPCertificate(parentCert)
 	chain := parent.Parent
 	for chain != nil && len(checkParent) > 0 {
 		key := parentCert.Certificate.AuthorityKeyId
 		upperCert, found := v.ValidObjects[string(key)]
 		if !found {
-			return errors.New(fmt.Sprintf("One of the parents (%x) of %x is not valid", key, ski))
+			//return errors.New(fmt.Sprintf("One of the parents (%x) of %x is not valid", key, ski))
+			return NewCertificateErrorParent(cert, parentCert, errors.New(fmt.Sprintf("ancestor %x is missing", key)))
 		}
-		chainCert, ok := upperCert.Resource.(*librpki.RPKI_Certificate)
+		chainCert, ok := upperCert.Resource.(*librpki.RPKICertificate)
 		if !ok {
-			return errors.New(fmt.Sprintf("One of the parents (%x) of %x is not a RPKI Certificate", key, ski))
+			//return errors.New(fmt.Sprintf("One of the parents (%x) of %x is not a RPKI Certificate", key, ski))
+			return NewCertificateErrorParent(cert, parentCert, errors.New(fmt.Sprintf("ancestor %x is not a rpki certificate", key)))
 		}
-		validsTmp, invalidsTmp, checkParentTmp := librpki.ValidateIPCertificateList(checkParent, chainCert)
-		valids = append(valids, validsTmp...)
-		invalids = append(invalids, invalidsTmp...)
+		validTmp, invalidTmp, checkParentTmp := librpki.ValidateIPCertificateList(checkParent, chainCert)
+		validIPs = append(validIPs, validTmp...)
+		invalidIPs = append(invalidIPs, invalidTmp...)
 		checkParent = checkParentTmp
 		chain = chain.Parent
 	}
-	if len(invalids) > 0 {
-		return errors.New(fmt.Sprintf("%x contains invalid IP addresses: %v", ski, invalids))
-	}
 
 	// Check ASNs
-	validsASN, invalidsASN, checkParentASN := cert.ValidateASNCertificate(parentCert)
+	validASNs, invalidASNs, checkParentASN := cert.ValidateASNCertificate(parentCert)
 	chain = parent.Parent
 	for chain != nil && len(checkParentASN) > 0 {
 		key := parentCert.Certificate.AuthorityKeyId
 		upperCert, found := v.ValidObjects[string(key)]
 		if !found {
-			return errors.New(fmt.Sprintf("One of the parents (%x) of %x is not valid", key, ski))
+			return NewCertificateErrorParent(cert, parentCert, errors.New(fmt.Sprintf("ancestor %x is not valid", key)))
 		}
-		chainCert, ok := upperCert.Resource.(*librpki.RPKI_Certificate)
+		chainCert, ok := upperCert.Resource.(*librpki.RPKICertificate)
 		if !ok {
-			return errors.New(fmt.Sprintf("One of the parents (%x) of %x is not a RPKI Certificate", key, ski))
+			return NewCertificateErrorParent(cert, parentCert, errors.New(fmt.Sprintf("ancestor %x is not a rpki certificate", key)))
 		}
-		validsTmp, invalidsTmp, checkParentTmp := librpki.ValidateASNCertificateList(checkParentASN, chainCert)
-		validsASN = append(validsASN, validsTmp...)
-		invalidsASN = append(invalidsASN, invalidsTmp...)
+		validTmp, invalidTmp, checkParentTmp := librpki.ValidateASNCertificateList(checkParentASN, chainCert)
+		validASNs = append(validASNs, validTmp...)
+		invalidASNs = append(invalidASNs, invalidTmp...)
 		checkParentASN = checkParentTmp
 		chain = chain.Parent
 	}
-	if len(invalidsASN) > 0 {
-		return errors.New(fmt.Sprintf("%x contains invalid ASNs: %v", ski, invalidsASN))
+
+	if len(invalidIPs) > 0 || len(invalidASNs) > 0 {
+		//return errors.New(fmt.Sprintf("%x contains invalid ASNs: %v", ski, invalidsASN))
+		//return errors.New(fmt.Sprintf("%x contains invalid IP addresses: %v", ski, invalids))
+		return NewCertificateErrorResource(cert, invalidIPs, invalidASNs)
 	}
 
 	return nil
 }
 
-func (v *Validator) AddROA(pkifile *PKIFile, roa *librpki.RPKI_ROA) (bool, *Resource, error) {
+func (v *Validator) AddROA(pkifile *PKIFile, roa *librpki.RPKIROA) (bool, *Resource, error) {
 	valid, _, res, err := v.AddCert(roa.Certificate, false)
 	if res == nil {
 		return valid, res, errors.New(fmt.Sprintf("Resource is empty: %v", err))
@@ -505,10 +576,16 @@ func (v *Validator) AddROA(pkifile *PKIFile, roa *librpki.RPKI_ROA) (bool, *Reso
 	}
 	v.ROA[string(key)] = res_roa
 
+	if err != nil {
+		errRes := NewResourceErrorWrap(roa, err)
+		errRes.InnerValidity = valid
+		err = errRes
+	}
+
 	return valid, res_roa, err
 }
 
-func (v *Validator) ValidateROA(roa *librpki.RPKI_ROA) error {
+func (v *Validator) ValidateROA(roa *librpki.RPKIROA) error {
 	err := roa.ValidateEntries()
 	if err != nil {
 		return errors.New(fmt.Sprintf("Could not validate certificate due to wrong entry: %v", err))
@@ -516,7 +593,7 @@ func (v *Validator) ValidateROA(roa *librpki.RPKI_ROA) error {
 	return nil
 }
 
-func (v *Validator) AddManifest(pkifile *PKIFile, mft *librpki.RPKI_Manifest) (bool, []*PKIFile, *Resource, error) {
+func (v *Validator) AddManifest(pkifile *PKIFile, mft *librpki.RPKIManifest) (bool, []*PKIFile, *Resource, error) {
 	pathCert := ExtractPathManifest(mft)
 
 	valid, _, res, err := v.AddCert(mft.Certificate, false)
@@ -542,6 +619,12 @@ func (v *Validator) AddManifest(pkifile *PKIFile, mft *librpki.RPKI_Manifest) (b
 	}
 	v.Manifest[string(key)] = res_mft
 
+	if err != nil {
+		errRes := NewResourceErrorWrap(mft, err)
+		errRes.InnerValidity = valid
+		err = errRes
+	}
+
 	return valid, pathCert, res_mft, err
 }
 
@@ -566,10 +649,10 @@ func (v *Validator) AddCRL(crl *pkix.CertificateList) (bool, *Resource, error) {
 		valid = true
 	}
 
-	var parentCert *librpki.RPKI_Certificate
+	var parentCert *librpki.RPKICertificate
 	if hasParent && valid {
 		var ok bool
-		parentCert, ok = parent.Resource.(*librpki.RPKI_Certificate)
+		parentCert, ok = parent.Resource.(*librpki.RPKICertificate)
 		if !ok {
 			valid = false
 		}
@@ -593,7 +676,7 @@ func (v *Validator) AddCRL(crl *pkix.CertificateList) (bool, *Resource, error) {
 				key := string(aki) + revoked.SerialNumber.String()
 				child, found := v.CertsSerial[key]
 				if found {
-					childConv := child.Resource.(*librpki.RPKI_Certificate)
+					childConv := child.Resource.(*librpki.RPKICertificate)
 					if childConv.Certificate.SerialNumber.Cmp(revoked.SerialNumber) == 0 {
 						v.InvalidateObject(childConv.Certificate.SubjectKeyId)
 					}
@@ -632,7 +715,7 @@ func DetermineType(path string) int {
 	return TYPE_UNKNOWN
 }
 
-func ExtractPathCert(cert *librpki.RPKI_Certificate) []*PKIFile {
+func ExtractPathCert(cert *librpki.RPKICertificate) []*PKIFile {
 	fileList := make([]*PKIFile, 0)
 
 	var repo string
@@ -665,14 +748,16 @@ func ExtractPathCert(cert *librpki.RPKI_Certificate) []*PKIFile {
 	return fileList
 }
 
-func ExtractPathManifest(mft *librpki.RPKI_Manifest) []*PKIFile {
+// Returns the list of files from the Manifest
+func ExtractPathManifest(mft *librpki.RPKIManifest) []*PKIFile {
 	fileList := make([]*PKIFile, 0)
 	for _, file := range mft.Content.FileList {
-		curFile := file.File
+		curFile := file.Name
 		path := string(curFile)
 		item := PKIFile{
-			Type: DetermineType(path),
-			Path: path,
+			Type:         DetermineType(path),
+			Path:         path,
+			ManifestHash: file.GetHash(),
 		}
 		fileList = append(fileList, &item)
 	}
@@ -683,18 +768,79 @@ func (sm *SimpleManager) AddInitial(fileList []*PKIFile) {
 	sm.PutFiles(fileList)
 }
 
+// Given a file, invalidates the certificate parent of the Manifest in which the file is listed in
+func (sm *SimpleManager) InvalidateManifestParent(file *PKIFile, mftError error) {
+	if file != nil && file.Parent != nil && file.Parent.Type == TYPE_MFT && file.Parent.Parent != nil && file.Parent.Parent.Type == TYPE_CER {
+		res, ok := sm.ResourceOfPath[file.Parent.Parent]
+
+		if ok && res != nil && res.Resource != nil {
+			cert, ok := res.Resource.(*librpki.RPKICertificate)
+			if ok {
+				sm.Validator.InvalidateObject(cert.Certificate.SubjectKeyId)
+
+				err := NewCertificateErrorManifestRevocation(cert, mftError, file.Parent, file)
+				sm.reportErrorFile(err, file.Parent.Parent, nil)
+
+			} else {
+				sm.Log.Debugf("Could not invalidate certificate because incorrect resource")
+			}
+		} else {
+			sm.Log.Debugf("Could not invalidate certificate because not found in list")
+		}
+	}
+}
+
+func (sm *SimpleManager) InvalidateCRLParent(file *PKIFile, crlError error) {
+	if file != nil && file.Parent != nil && file.Parent.Type == TYPE_CRL && file.Parent.Parent != nil && file.Parent.Parent.Type == TYPE_CER {
+		res, ok := sm.Validator.ObjectsPath[file.Parent.Parent.Path]
+
+		if ok && res != nil && res.Resource != nil {
+			cert, ok := res.Resource.(*librpki.RPKICertificate)
+			if ok {
+				sm.Validator.InvalidateObject(cert.Certificate.SubjectKeyId)
+
+				err := NewCertificateErrorCRLRevocation(cert, crlError, file.Parent, file)
+				sm.reportErrorFile(err, file.Parent.Parent, nil)
+
+			} else {
+				sm.Log.Debugf("Could not invalidate certificate because incorrect resource")
+			}
+		} else {
+			sm.Log.Debugf("Could not invalidate certificate because not found in list")
+		}
+	}
+}
+
 func (sm *SimpleManager) ExploreAdd(file *PKIFile, data *SeekFile, addInvalidChilds bool) {
 	sm.Explored[file.ComputePath()] = true
 	valid, subFiles, res, err := sm.Validator.AddResource(file, data.Data)
+
+	/*if !valid || err != nil {
+		if sm.StrictManifests {
+			// will also invalidate when ROA is expired
+			sm.InvalidateManifestParent(file, err)
+		}
+	}*/
+
 	if err != nil {
+		//sm.InvalidateCRLParent(file, err)
+
+		switch err.(type) {
+		case *FileError:
+		case *ResourceError:
+		case *CertificateError:
+		default:
+			fe := NewFileError(err)
+			fe.AddFileErrorInfo(file, data)
+			err = fe
+		}
+
 		if sm.Log != nil {
-			sm.Log.Errorf("Error adding Resource %v: %v", file.Path, err)
+			sm.reportErrorFile(err, file, data)
 		}
 	}
 	if !valid && err == nil {
-		if sm.Log != nil {
-			sm.Log.Warnf("Resource %v is invalid: %v", file.Path, err)
-		}
+		sm.reportErrorFile(err, file, data)
 	}
 	for _, subFile := range subFiles {
 		subFile.Parent = file
@@ -702,6 +848,7 @@ func (sm *SimpleManager) ExploreAdd(file *PKIFile, data *SeekFile, addInvalidChi
 	if addInvalidChilds || valid {
 		sm.PutFiles(subFiles)
 		sm.PathOfResource[res] = file
+		sm.ResourceOfPath[file] = res
 	}
 }
 
@@ -718,23 +865,39 @@ func (sm *SimpleManager) Explore(notMFT bool, addInvalidChilds bool) int {
 
 		file, hasMore, err = sm.GetNextExplore()
 		if err != nil {
-			if sm.Log != nil {
-				sm.Log.Errorf("Error getting file: %v", err)
-			}
+			sm.reportError(err)
 		} else {
 			count++
 		}
 		if !notMFT || file.Type != TYPE_MFT {
 			data, err := sm.GetNextFile(file)
 
-			if err != nil {
-				if sm.Log != nil {
-					sm.Log.Errorf("Error exploring file: %v", err)
+			if err == nil && data != nil && sm.StrictHash && data.Sha256 != nil && file.ManifestHash != nil {
+				if bytes.Compare(data.Sha256, file.ManifestHash) != 0 {
+					errHash := NewResourceErrorHash(data.Sha256, file.ManifestHash)
+					errHash.AddFileErrorInfo(file, data)
+					err = errHash
 				}
+			}
+
+			if err != nil || data == nil {
+
+				// This invalidates the Manifests' CA when a file is not found
+				if sm.StrictManifests {
+					sm.InvalidateManifestParent(file, nil)
+					//sm.reportErrorFile(err, file, data)
+				}
+
+			}
+
+			if err != nil {
+				sm.reportErrorFile(err, file, data)
+
+				sm.InvalidateCRLParent(file, err)
 			} else if data != nil {
 				sm.ExploreAdd(file, data, addInvalidChilds)
 				hasMore = sm.HasMore()
-			} else {
+			} else { // data == nil && err == nil -> file was not found
 				if sm.Log != nil {
 					sm.Log.Debugf("GetNextFile returned nothing")
 				}
@@ -743,9 +906,7 @@ func (sm *SimpleManager) Explore(notMFT bool, addInvalidChilds bool) int {
 			err = sm.GetNextRepository(file, sm.ExploreAdd)
 			sm.Explored[file.Repo] = true
 			if err != nil {
-				if sm.Log != nil {
-					sm.Log.Errorf("Error exploring repository: %v", err)
-				}
+				sm.reportErrorFile(err, file, nil)
 			}
 		}
 		hasMore = sm.HasMore()
